import {desiredHeight, desiredWidth, getScreen} from "./screen.ts";
import {BG_OUTER} from "./colors.ts";
import {AlignX, mainFont} from "./font.ts";
import {getInput} from "./input.ts";

type Point = {x: number, y: number}

class MenuCamera {
  // measured in whole screens
  position: Point;
  target: Point

  constructor({position, target}: {position: Point, target: Point}) {
    this.position = position;
    this.target = target;
  }

  update() {
    let adjust = (x0: number, x1: number) => {
      if (Math.abs(x1 - x0) < 0.01) { return x1; }
      return (x0 * 8 + x1 * 2) / 10;
    }
    this.position.x = adjust(this.position.x, this.target.x);
    this.position.y = adjust(this.position.y, this.target.y);
  }
}

type GameState = "Overview" | "Gameplay" | "Thralls";

function getScreenLocation(state: GameState): {x: number, y: number} {
  if (state === "Overview") {
    return {x: 0.0, y: 0.0}
  }
  if (state === "Gameplay") {
    return {x: 1.0, y: 0.0}
  }
  if (state === "Thralls") {
    return {x: 0.0, y: 1.0}
  }

  throw `invalid state: ${state}`
}

export class Game {
  camera: MenuCamera;
  state: GameState;

  constructor() {
    this.camera = new MenuCamera({
      position: {x: 0.0, y: 0.0},
      target: {x: 0.0, y: 0.0}
    });
    this.state = "Overview";
  }

  update() {
    if (getInput().isPressed("a") || getInput().isPressed("w")) {
      this.state = "Overview"
    }
    if (getInput().isPressed("d")) {
      this.state = "Gameplay"
    }
    if (getInput().isPressed("s")) {
      this.state = "Thralls"
    }

    this.camera.target = getScreenLocation(this.state);
    this.camera.update();

    // state-specific updates
    this.updateOverview();
  }

  draw() {
    let screen = getScreen();
    let ctx = screen.makeContext();

    // draw screen background
    ctx.fillStyle = BG_OUTER;
    ctx.fillRect(0, 0, screen.w, screen.h);

    this.drawOverview();
    // this.drawGameplay();

    // we draw all states at once and pan between them
    // mainFont.drawText({ctx: ctx, text: "You have been given a gift.", x: 0, y: 0})
    let xy = this.getCameraMouseXy();
    if (xy != null) {
      ctx = this.makeCameraContext();
      ctx.globalCompositeOperation = "difference";
      if (getInput().isDown("leftMouse")) {
        ctx.fillStyle = "#ff0";
      } else {
        ctx.fillStyle = "#fff";
      }
      ctx.fillRect(xy.x - 1, xy.y - 1, 3, 3);
      ctx.fillRect(xy.x, xy.y - 1, 1, 3);
      ctx.fillRect(xy.x - 1, xy.y, 3, 1);
    }
  }

  getCameraOffset(): Point {
    let screen = getScreen();
    let {w, h} = screen;
    return {
      x: Math.floor(w * this.camera.position.x),
      y: Math.floor(h * this.camera.position.y)
    };
  }

  getCameraMouseXy(): Point | null {
    let xy = getInput().mouseXy();
    if (xy == null) {
      return null;
    }
    let {x: dx, y: dy} = this.getCameraOffset();
    return {
      x: xy.x + dx,
      y: xy.y + dy,
    };
  }

  makeCameraContext() {
    let screen = getScreen();
    let ctx = screen.makeContext();
    let {x, y} = this.getCameraOffset();

    ctx.translate(-x, -y);
    return ctx;
  }

  getPaneRegionForGameState(gameState: GameState) {
    let screen = getScreen();
    let {w, h} = screen;
    let overallScreenLocation = getScreenLocation(gameState);

    let bigPaneX = overallScreenLocation.x * w;
    let bigPaneY = overallScreenLocation.y * h;
    let bigPaneW = w;
    let bigPaneH = h;

    let smallPaneW = desiredWidth;
    let smallPaneH = desiredHeight;
    let smallPaneX = Math.floor(bigPaneX + (bigPaneW - smallPaneW) / 2)
    let smallPaneY = Math.floor(bigPaneY + (bigPaneH - smallPaneH) / 2)

    return {
      big: {
        position: {x: bigPaneX, y: bigPaneY},
        size: {x: bigPaneW, y: bigPaneH}
      },
      small: {
        position: {x: smallPaneX, y: smallPaneY},
        size: {x: smallPaneW, y: smallPaneH}
      }
    }

  }

  updateOverview() {

  }

  drawOverview() {
    let region = this.getPaneRegionForGameState("Overview")
    let ctx = this.makeCameraContext()
    ctx.translate(region.small.position.x, region.small.position.y)

    // ctx.strokeStyle = "#ffffff";
    // ctx.strokeRect(0.5, 0.5, region.small.size.x - 1, region.small.size.y - 1);

    // draw skills menu
    for (let [i, skillName] of [
      "Screech", "Stop Talking", "Flatter", "Dazzle", "Chug", "Respect Elders"
    ].entries()) {
      mainFont.drawText({ctx: ctx, text: skillName, x: 4, y: i * 16, alignX: AlignX.Left})
      mainFont.drawText({ctx: ctx, text: "150", x: 160 - 4, y: i * 16, alignX: AlignX.Right, color: "#ff0"})
    }

    mainFont.drawText({ctx: ctx, text: "Stop embarrassing yourself\naround other immortals.\nKeep to yourself and they\nwon't make fun of you.\n\nCost reduced by INT and AGI", x: 160 + 4, y: 0, forceWidth: 384 - 160 - 8})

    ctx.globalCompositeOperation = "difference";
    ctx.fillStyle = "#fff"
    ctx.fillRect(0, 16, 160, 16);
    ctx.fillRect(160, 0, 384 - 160, 96);

    // draw character profile
    ctx = this.makeCameraContext();
    ctx.translate(region.small.position.x, region.small.position.y)
    // ctx.fillStyle = "#200";
    // ctx.fillRect(0, 96, 384, 192)
    // ctx.fillStyle = "#ff0"; ctx.fillRect(176, 176, 32, 16);
    // draw ascii art
    mainFont.drawText(
      {ctx, x: 272, y: 160, text: "       \n; 0  0 \n \\    / \n  \\\\'  \n /      \n .'  '\n/_'     ", color: "#334", alignX: AlignX.Center}
    )
    mainFont.drawText(
      {ctx, x: 272, y: 160, text: "      \n  0  0  \n   . . ", color: "#fff", alignX: AlignX.Center}
    )
    mainFont.drawText(
      {ctx, x: 272, y: 160, text: "   ^  ^ \n<; 0  0   ", color: "#448", alignX: AlignX.Center}
    )
    mainFont.drawText(
      {ctx, x: 272, y: 160, text: "      \n  0  0  ", color: "#f00", alignX: AlignX.Center}
    )

    // draw character name
    mainFont.drawText({ctx, x: 4 + 32, y: 96 + 16 + 0, text: "Pyrex"})
    mainFont.drawText({ctx, x: 4 + 32, y: 96 + 16 + 16, text: "Level 6", color: "#ccc"})

    // draw attributes
    for (let [i, {name, value}] of [
      {name: "AGI", value: 10},
      {name: "INT", value: 20},
      {name: "CHA", value: 30},
      {name: "PSI", value: 40},
    ].entries()) {
      mainFont.drawText({ctx, x: 32 + 4, y: 96 + 64 + 16 * i, text: name, color: "#fff"})
      mainFont.drawText({ctx, x: 32 + 4 + 32, y: 96 + 64 + 16 * i, text: "" + value, color: "#ccc"})
    }

    mainFont.drawText({ctx, x: 32 + 4, y: 96 + 144, text: "MN ", color: "#fff"})
    mainFont.drawText({ctx, x: 32 + 4 + 24, y: 96 + 144, text: "$400", color: "#ccc"})
    mainFont.drawText({ctx, x: 32 + 4, y: 96 + 160, text: "BLD", color: "#fff"})
    mainFont.drawText({ctx, x: 32 + 4 + 32, y: 96 + 160, text: "10cc", color: "#ccc"})

    // draw character thoughts
    // mainFont.drawText({ctx, x: 256, y: 96 + 16, text: "[I'm peckish!]\n[I must feed.]", alignX: AlignX.Center})
    mainFont.drawText({ctx, x: 272, y: 96 + 16, text: "[Learn Stop Talking?]", alignX: AlignX.Center})
    mainFont.drawText({ctx, x: 272, y: 96 + 32, text: "[Yes] [No]", alignX: AlignX.Center})

    // draw thralls
    ctx.fillStyle = "#f00"
    ctx.fillRect(0, 288, 128, 96);
    ctx.fillStyle = "#0f0"
    ctx.fillRect(128, 288, 128, 96);
    ctx.fillStyle = "#00f"
    ctx.fillRect(256, 288, 128, 96);
  }
}

export let game = new Game();
