import {getScreen} from "./screen.ts";

function handleKey(e: KeyboardEvent, down: boolean) {
  active.handleDown(e.key, down);
}
function handleMouseOut() {
  active.handleMouseMove(-1, -1);
}

function handleMouseMove(canvas: HTMLCanvasElement, m: MouseEvent) {
  if (canvas.offsetWidth == 0 || canvas.offsetHeight == 0) {
    return;
  }
  active.handleMouseMove(m.offsetX / canvas.offsetWidth, m.offsetY / canvas.offsetHeight);
}

function handleMouseButton(canvas: HTMLCanvasElement, m: MouseEvent, down: boolean) {
  if (canvas.offsetWidth == 0 || canvas.offsetHeight == 0) {
    return;
  }
  active.handleMouseMove(m.offsetX / canvas.offsetWidth, m.offsetY / canvas.offsetHeight);
  let button: KnownButton | null = (
    m.button == 0 ? "leftMouse" :
      m.button == 1 ? "rightMouse" :
        null
  )
  if (button != null) {
    active.handleDown(button, down);
  }
}


export function setupInput(canvas: HTMLCanvasElement) {
  canvas.addEventListener("keyup", (k) => handleKey(k, false));
  document.addEventListener("keyup", (k) => handleKey(k, false));
  canvas.addEventListener("keydown", (k) => handleKey(k, true));
  document.addEventListener("keydown", (k) => handleKey(k, true));
  canvas.addEventListener("mouseout", (_) => handleMouseOut());
  canvas.addEventListener("mousemove", (m) => handleMouseMove(canvas, m));
  canvas.addEventListener("mousedown", (m) => handleMouseButton(canvas, m, true));
  canvas.addEventListener("mouseup", (m) => handleMouseButton(canvas, m, false));
}

type KnownKey = "w" | "a" | "s" | "d";
type KnownButton = "leftMouse" | "rightMouse";

class Input {
  #down: Record<string, boolean>;
  #previousDown: Record<string, boolean>;
  #mouseXy: {x: number, y: number} | null;

  constructor() {
    this.#down = {};
    this.#previousDown = {};
    this.#mouseXy = null;
  }

  update() {
    this.#previousDown = {...this.#down};
  }

  handleDown(name: string, down: boolean) {
    if (down) {
      this.#down[name] = down;
    }
    else {
      delete this.#down[name];
    }
  }

  handleMouseMove(x: number, y: number) {
    let screen = getScreen();
    if (x < 0.0 || x >= 1.0) { this.#mouseXy = null; }
    if (y < 0.0 || y >= 1.0) { this.#mouseXy = null; }

    let w = screen.w;
    let h = screen.h;
    this.#mouseXy = {
      x: Math.floor(x * w),
      y: Math.floor(y * h),
    }
  }

  isDown(key: KnownKey | KnownButton) : boolean {
    return this.#down[key];
  }

  isPressed(key: KnownKey | KnownButton) : boolean {
    return this.#down[key] && !this.#previousDown[key];
  }

  isReleased(key: KnownKey | KnownButton) : boolean {
    return !this.#down[key] && this.#previousDown[key];
  }

  mouseXy(): {x: number, y: number} | null {
    if (this.#mouseXy == null) {
      return null;
    }
    return {...this.#mouseXy}
  }
}

let active = new Input();

export function getInput(): Input {
  return active;
}